import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import java.io.*;
import java.util.*;

public class Dictionary extends MIDlet implements CommandListener  
{ 
    TextBox mainMenu;
    static Command QUIT_CMD = new Command(Locale.EXIT, Command.EXIT, 2);
    static Command BACK_CMD = new Command(Locale.BACK, Command.BACK, 2);
    static Command TRANSLATE_CMD = new Command(Locale.TRANSLATE, Command.OK, 1);
    
    static final int ENG = 0;
    static final int RUS = 1;

    static final String LANGUAGE[] = {"eng", "rus"};

    static final int MAX_LENGTH = 32;
    static final int MAX_FILE_SIZE = 8*1024;
    static final int SELECTION_LIMIT = 32;

    class Volume { 
        String last;
        int    index;
    }

    Volume[][] volumes;
    int[]      nVolumes;
    byte[]     buf = new byte[MAX_FILE_SIZE];

    int findVolume(int lang, String word) { 
        int l = 0, r = nVolumes[lang];
        while (l < r) { 
            int m = (l + r) >> 1;
            if (word.compareTo(volumes[lang][m].last) > 0) { 
                l = m + 1;
            } else { 
                r = m;
            }
        }
        return r;
    }
    
    int findVolume(int lang, int index) { 
        int l = 0, r = nVolumes[lang];
        while (l < r) { 
            int m = (l + r) >> 1;
            if (index > volumes[lang][m].index) { 
                l = m + 1;
            } else { 
                r = m;
            }
        }
        // System.out.println("Found word with index " + index + " in volume " +  r + ", total volumes " + nVolumes[lang]);
        return r;
    }
    
        

    public Dictionary () throws IOException
    {
        volumes = new Volume[2][];
        nVolumes = new int[2];
        buildIndex(ENG, "/eng.idx");
        buildIndex(RUS, "/rus.idx");
        mainMenu = new TextBox(Locale.PREFIX, "", MAX_LENGTH, TextField.ANY);
        mainMenu.addCommand(QUIT_CMD);
        mainMenu.addCommand(TRANSLATE_CMD);
        mainMenu.setCommandListener(this) ;
    }


    protected void destroyApp(boolean unconditional) 
    {
        mainMenu = null;
    }    

    protected  void pauseApp() 
    {
    }

    protected void startApp() 
    {
        Display.getDisplay(this).setCurrent(mainMenu);
    }

    static int unpackInt(byte[] arr, int offs) { 
        return (arr[offs] << 24) | ((arr[offs+1] & 0xFF) << 16) | ((arr[offs+2] & 0xFF) << 8) | (arr[offs+3] & 0xFF);
    }
        

    static String unpackString(byte[] bytes, int offs, int len, int lang) { 
        char[] chars = new char[len];
        for (int i = 0; i < len; i++) { 
            chars[i] = (char)(bytes[offs + i] & 0xFF);
        }
        return new String(chars);
        /*
        if (lang == RUS) {
            char[] chars = new char[len];
            for (int i = 0; i < len; i++) { 
                byte b = bytes[offs + i];
                chars[i] = (char)(b < 0 ? b == -88 ? 0x401 : b == -72 ? 0x451 : 
                                  (0x400 | ((b & 0x7f) - 0x30)) : b);
            }
            return new String(chars);
        } else { 
            try { 
                return new String(bytes, offs, len, "ISO8859-2");
            } catch (UnsupportedEncodingException x) { 
                System.out.println("Ecoding not suppoted");
                return new String(bytes, offs, len);
            }
        }
        */
    }

    static String unpackString(byte[] bytes, int lang) { 
        return unpackString(bytes, 0, bytes.length, lang);
    }

    private void buildIndex(int lang, String resource) throws IOException 
    {
        DataInputStream in = new DataInputStream(getClass().getResourceAsStream(resource));
        Vector volumesList = new Vector();
        try { 
            while (true) { 
                int len = in.readByte();
                byte[] bytes = new byte[len];
                in.read(bytes);
                Volume v = new Volume();
                v.last = unpackString(bytes, lang);
                v.index = in.readInt();
                volumesList.addElement(v);
            }
        } catch (EOFException x) {}
        in.close();

        int size = volumesList.size();
        volumes[lang] = new Volume[size];
        for (int i = 0; i < size; i++) { 
            volumes[lang][i] = (Volume)volumesList.elementAt(i);
        }
        nVolumes[lang] = size;
    }

    public String getTranslation(Translation trans) { 
        try { 
            int lang = RUS - trans.lang;       
            StringBuffer sb = new StringBuffer();
            for (int i = 0; i < trans.index.length; i++) { 
                int index = trans.index[i];
                int v = findVolume(lang, index);
                InputStream in = getClass().getResourceAsStream(LANGUAGE[lang] + '.' + (v+1));
                int size = in.read(buf);
                in.close();
                String prevWord = null;
                if (v > 0) { 
                    index -= volumes[lang][v-1].index + 1;
                    prevWord = volumes[lang][v-1].last;
                }
                int offs = 0;
                do { 
                    int len = buf[offs++] & 0xFF;
                    int firstChar = buf[offs] & 0xFF;
                    String word = unpackString(buf, offs, len, lang);
                    offs += len+4;
                    // System.out.println("v=" + v + ", firstChar=" + firstChar + ", offs=" + offs + ", len=" + len + ", word=" + word + ", prevWord=" + prevWord);
                    if (len > 0 && firstChar <= ' ') {
                        word = prevWord.substring(0, firstChar) + word.substring(1);
                    }
                    prevWord = word;
                } while (--index >= 0);
                
                if (trans.index.length == 1) { 
                    return prevWord;
                } else {
                    sb.append(Integer.toString(i+1));
                    sb.append(". ");
                    sb.append(prevWord);
                    sb.append('\n');
                }
            }
            return sb.toString();
	} catch (IOException x)   { 
	    x.printStackTrace();
	}
        return null;
    }          


    public Vector getSelection(String prefix) 
    { 
        int lang = ENG;
        /*
        for (int i = prefix.length(); --i >= 0;) { 
            if ((prefix.charAt(i) & 0xFF00) == 0x400) { 
                lang = RUS;
                break;
            }
        }
        */
        Vector v = getSelection(prefix, lang);
        if (v.size() == 0 || !((Translation)v.elementAt(0)).word.startsWith(prefix)) { 
            Vector v2 = getSelection(prefix, RUS - lang);  
            if (v2.size() != 0 && ((Translation)v2.elementAt(0)).word.startsWith(prefix)) { 
                v = v2;
            }
        }
        return v;
    }

    private Vector getSelection(String prefix, int lang) 
    { 
        Vector selection = new Vector();
        try { 
            Translation trans = null;
            for (int v = findVolume(lang, prefix); v < nVolumes[lang]; v++) { 
                InputStream in = getClass().getResourceAsStream(LANGUAGE[lang] + '.' + (v+1));
                int size = in.read(buf);
                in.close();
                String prevWord = v > 0 ? volumes[lang][v-1].last : null;
                int offs = 0;
                boolean found = false;
                while (offs < size) { 
                    int len = buf[offs++] & 0xFF;
                    int firstChar = buf[offs] & 0xFF;
                    String word = unpackString(buf, offs, len, lang);
                    offs += len;
                    int index = unpackInt(buf, offs);
                    offs += 4;
                    if (len > 0 && firstChar <= ' ') {
                        word = prevWord.substring(0, firstChar) + word.substring(1);
                    }
                    if (word.compareTo(prefix) >= 0) {                         
                        if (word.startsWith(prefix)) { 
                            found = true;
                        } else if (found) { 
                            return selection;
                        }
                        if (word.equals(prevWord)) { 
                            int[] indices = new int[trans.index.length+1];
                            System.arraycopy(trans.index, 0, indices, 0, trans.index.length);
                            indices[indices.length-1] = index;
                            trans.index = indices;
                        } else {                                       
                            if (selection.size() == SELECTION_LIMIT) { 
                                return selection;
                            }
                            trans = new Translation(word, new int[]{index}, lang);
                            selection.addElement(trans);
                        }
                    }
                    prevWord = word;
                }
            }
	} catch (IOException x)   { 
	    x.printStackTrace();
	}
        return selection;
    }
    
    public void commandAction(Command c, Displayable d) 
    {
        if (c == QUIT_CMD) 
        {
            destroyApp(true);
            notifyDestroyed();
        } 
        else 
        {
            String key = mainMenu.getString().toLowerCase();
            if (key.length() != 0) 
            {
                mainMenu.setTitle(Locale.ONE_MINUTE);
                mainMenu.setString(Locale.SEARCHING);
                Vector selection = getSelection(key);
                switch (selection.size()) { 
                case 0:
                    Alert a = new Alert(key, Locale.NO_VARIANTS, null, AlertType.ERROR);
                    a.setTimeout(Alert.FOREVER);
                    Display.getDisplay(this).setCurrent(a, mainMenu);
                    break;
                case 1:
                    Translation t = (Translation)selection.firstElement();
                    if (t.word.equals(key)) { 
                        new TranslationBox(this, mainMenu, (Translation)selection.firstElement());
                        break;
                    }
                default:
                    new PrefixList(this, mainMenu, selection);
                    break;
                }
                mainMenu.setTitle(Locale.PREFIX);
                mainMenu.setString(key);
            }
        }
    }
}



